import lightning
import torch
import torchmetrics

from experiments.metrics import default_classification_metrics
from sgu_mlp import SGUMLPMixer


class LitSGUMLPMixer(lightning.LightningModule):
    def __init__(
        self,
        model_params,
        optimizer_params,
        metrics=None,
        criterion=None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.save_hyperparameters(ignore=["metrics", "criterion"])

        if metrics is None:
            metrics = default_classification_metrics(self.hparams.model_params["num_classes"])
        self.model = SGUMLPMixer(**model_params)
        if criterion is None:
            self.criterion = torch.nn.CrossEntropyLoss()
        else:
            self.criterion = criterion
        self.optimizer_cls = torch.optim.AdamW

        self.train_metrics = torchmetrics.MetricCollection(
            dict(
                metrics.get("train", default_classification_metrics(self.model.num_classes)["train"])
            )
        )
        self.test_metrics = torchmetrics.MetricCollection(
            dict(metrics.get("test", default_classification_metrics(self.model.num_classes)["test"]))
        )

    def forward(self, x):
        return self.model(x)

    def _step(self, batch, batch_idx, split, on_step=False, on_epoch=True):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.criterion(y_hat, y)

        self.log(f"{split}_loss", loss, on_step=True, on_epoch=True, prog_bar=True)
        metrics = self.train_metrics if split == "train" else self.test_metrics
        for name, metric in metrics.items():
            self.log(
                f"{split}_{name}",
                metric(y_hat, y),
                on_step=on_step,
                on_epoch=on_epoch,
                prog_bar=True,
            )
        return loss

    def training_step(self, batch, batch_idx):
        return self._step(batch, batch_idx, "train", on_step=True, on_epoch=False)

    def validation_step(self, batch, batch_idx):
        with torch.no_grad():
            return self._step(batch, batch_idx, "val", on_step=False, on_epoch=True)

    def test_step(self, batch, batch_idx):
        with torch.no_grad():
            return self._step(batch, batch_idx, "test", on_step=False, on_epoch=True)

    def configure_optimizers(self):
        return self.optimizer_cls(
            self.model.parameters(), **self.hparams.optimizer_params
        )



import dataclasses
import json
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from os import PathLike
from typing import Literal, Mapping, Sequence, Any

Split = Literal["train", "test", "validation"]
Metric = Literal["accuracy", "precision", "recall", "f1_score"]
TrainType = Literal["train-test", "cv"]


@dataclass
class JSONSerializableMixin(ABC):
    def to_dict(self) -> Mapping[str, Any]:
        return dataclasses.asdict(self)

    def to_json(self, filepath: PathLike = None) -> str:
        data = self.to_dict()
        if filepath is None:
            return json.dumps(data)
        else:
            with open(filepath, "w") as f:
                json.dump(data, f)
                return str(filepath)

    @classmethod
    def from_json(cls, filepath: PathLike):
        with open(filepath, "r") as f:
            return cls(**json.load(f))


@dataclass
class DatasetConfig(JSONSerializableMixin):
    name: str
    base_dir: str
    feature_files: list[str]
    labels_file: str
    labels_file_test: str | None = None
    na_label: int | None = None
    preprocessing: list | None = None

    @classmethod
    def from_json(cls, path):
        with open(path) as f:
            return cls(**json.load(f))


@dataclass
class TrainingConfig(JSONSerializableMixin):
    seed: int = 42
    type: TrainType = "cv"
    size: int = 5
    batch_size: int = 256
    epochs: int = 100
    early_stopping: bool = False


@dataclass
class ModuleConfig(JSONSerializableMixin):
    class_name: str
    args: Mapping[str, any]


@dataclass
class MetricsConfig(JSONSerializableMixin):
    task: Literal["classification", "regression"] = "classification"
    task_type: Literal["multiclass", "binary"] = "multiclass"
    num_classes: int = -1
    train: Sequence[Metric] = dataclasses.field(default_factory=lambda: ["accuracy"])
    test: Sequence[Metric] = dataclasses.field(
        default_factory=lambda: ["accuracy", "precision", "recall", "f1_score"]
    )


@dataclass
class ExperimentConfig(JSONSerializableMixin):
    datasets: Mapping[Split, Sequence[DatasetConfig]]
    model: ModuleConfig
    optimizer: ModuleConfig
    version: str = "0.0.1"
    run_id: str = "auto"
    name: str | None = None
    training: TrainingConfig = TrainingConfig()
    metrics: MetricsConfig = MetricsConfig()

    def __post_init__(self):
        if self.run_id == "auto":
            self.run_id = self._run_id()
        if self.name is None:
            self.name = self.run_id

    def _run_id(self):
        return f"{self.name}__{datetime.now().isoformat()}"

    def dataset(self, split: Split):
        dataset = self.datasets[split]
        if not dataset:
            dataset = self.datasets["train"]
        return dataset



from typing import Literal

import torch
import torchmetrics


class StableProbabilityMixin:
    def preprocess_inputs(self, preds, target, epsilon=1e-8):
        target = torch.nan_to_num(target, 0.0)
        target = target + epsilon
        target = target / target.sum(1, keepdim=True)

        preds_proba = torch.softmax(preds, 1) + epsilon
        preds_proba = preds_proba / preds_proba.sum(1, keepdim=True)

        return preds_proba, target


class StableCosineSimilarity(StableProbabilityMixin, torchmetrics.CosineSimilarity):
    def __init__(self, epsilon=1e-8, **kwargs):
        super().__init__(**kwargs)
        self.epsilon = epsilon

    def update(self, preds, target):
        preds, target = self.preprocess_inputs(preds, target, self.epsilon)
        super().update(preds, target)


class StableKLDivergence(StableProbabilityMixin, torchmetrics.KLDivergence):
    def __init__(self, epsilon=1e-8, **kwargs):
        super().__init__(**kwargs)
        self.log_prob = False
        self.epsilon = epsilon

    def update(self, preds, target):
        preds, target = self.preprocess_inputs(preds, target, self.epsilon)
        super().update(preds, target)


def js_divergence(preds, target, log_prob=False, reduction: Literal["mean", "sum"]="mean", normalize=True):
    m = 0.5 * (preds + target)
    kld_p = torchmetrics.functional.kl_divergence(preds, m, log_prob=log_prob, reduction=reduction)
    kld_q = torchmetrics.functional.kl_divergence(target, m, log_prob=log_prob, reduction=reduction)
    jsd = (kld_p + kld_q) * 0.5

    if normalize:
        jsd = jsd / torch.log(torch.tensor(2.0))

    return jsd


class StableJensonShannonDivergence(StableProbabilityMixin, torchmetrics.Metric):
    def __init__(self, epsilon=1e-8, reduction: Literal["mean", "sum"]="mean", normalize=True, **kwargs):
        super().__init__(**kwargs)
        self.reduction = reduction
        self.epsilon = epsilon
        self.normalize = normalize
        self.add_state("measures", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0, dtype=torch.int64), dist_reduce_fx="sum")

    def update(self, preds, target):
        preds, target = self.preprocess_inputs(preds, target, self.epsilon)
        self.measures += js_divergence(preds, target, False, "sum", self.normalize)
        self.total += target.shape[0]

    def compute(self):
        if self.reduction == "mean":
            return self.measures / self.total
        else:
            return self.measures


def hellinger_distance(preds, target, reduction: Literal["mean", "sum"] = "mean"):
    h = torch.sqrt(torch.sum((torch.sqrt(preds) - torch.sqrt(target))**2, -1)) * (1 / 2**0.5)
    return h.mean() if reduction == "mean" else h.sum()


class StableHellingerDistance(StableProbabilityMixin, torchmetrics.Metric):
    def __init__(self, epsilon=1e-8, reduction: Literal["mean", "sum"]="mean", normalize=True, **kwargs):
        super().__init__(**kwargs)
        self.reduction = reduction
        self.epsilon = epsilon
        self.normalize = normalize
        self.add_state("measures", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0, dtype=torch.int64), dist_reduce_fx="sum")

    def update(self, preds, target):
        preds, target = self.preprocess_inputs(preds, target, self.epsilon)
        self.measures += hellinger_distance(preds, target, reduction="sum")
        self.total += preds.shape[0]

    def compute(self):
        if self.reduction == "mean":
            return self.measures / self.total
        else:
            return self.measures


def get_metrics(num_labels: int):
    task = "binary" if num_labels == 2 else "multiclass"
    metric_args = dict(task=task, num_classes=num_labels)
    return dict(
        train=dict(
            accuracy=torchmetrics.Accuracy(**metric_args),
        ),
        test=dict(
            accuracy=torchmetrics.Accuracy(**metric_args),
            precision=torchmetrics.Precision(**metric_args),
            recall=torchmetrics.Recall(**metric_args),
            f1=torchmetrics.F1Score(**metric_args),
        ),
    )


def default_classification_metrics(num_classes):
    task = "binary" if num_classes == 2 else "multiclass"
    return dict(
        train=dict(
            accuracy=torchmetrics.Accuracy(task=task, num_classes=num_classes),
        ),
        test=dict(
            accuracy=torchmetrics.Accuracy(task=task, num_classes=num_classes),
        ),
    )



from argparse import ArgumentParser
from pathlib import Path

from .experiments import cv_experiment, ood_experiment
from .data import download_benchmark_datasets

def get_data_download_parser(subparsers):
    parser = subparsers.add_parser('download')
    parser.add_argument("--out_dir", type=str, default="data",
                       help="Path to the root directory of the datasets")
    parser.add_argument("--overwrite", action="store_true",
                       help="Overwrite existing datasets", default=False)
    return parser

def get_experiment_parser(subparsers):
    parser = subparsers.add_parser('experiment')
    parser.add_argument("name", choices=["replication"])
    parser.add_argument("--root_dir", type=str, default="data/config")
    return parser

def get_run_parser(subparsers):
    parser = subparsers.add_parser('run')
    parser.add_argument("type", choices=["cv", "ood"])
    parser.add_argument("cfg_path", type=str)
    return parser

if __name__ == '__main__':
    parser = ArgumentParser()

    subparsers = parser.add_subparsers(dest='command', required=True)
    get_data_download_parser(subparsers)
    get_experiment_parser(subparsers)
    get_run_parser(subparsers)

    args = parser.parse_args()

    if args.command == "download":
        download_benchmark_datasets(args.out_dir, overwrite=args.overwrite)

    if args.command == "experiment":
        cfg_path = Path(args.root_dir) / f"{args.name}.experiment.json"
        cv_experiment(str(cfg_path))

    if args.command == "run":
        match args.type:
            case "cv":
                cv_experiment(args.cfg_path)
            case "ood":
                ood_experiment(args.cfg_path)


import lightning
import torch
from sklearn.model_selection import KFold

from .data import get_dataloader
from .models import LitSGUMLPMixer


def run_train_test(
    dataloader_train: torch.utils.data.DataLoader,
    model_args: dict,
    trainer_args: dict,
    optimizer_args: dict,
    metrics: dict,
    meta_data: dict,
    dataloader_val: torch.utils.data.DataLoader = None,
    dataloader_test: torch.utils.data.Dataset | list = None,
    model_class=LitSGUMLPMixer,
    criterion=None,
):
    model = model_class(
        model_args, optimizer_args, metrics, criterion=criterion, meta_data=meta_data
    )
    trainer = lightning.Trainer(**trainer_args)
    trainer.fit(model, dataloader_train, dataloader_val)
    if dataloader_test is None:
        return
    if type(dataloader_test) is list:
        for dl in dataloader_test:
            trainer.test(model, dl)
    else:
        trainer.test(model, dataloader_test)


def run_cv(
    dataset_train: torch.utils.data.Dataset,
    model_args: dict,
    trainer_args: dict,
    optimizer_args: dict,
    metrics: dict,
    meta_data: dict,
    dataset_test: torch.utils.data.Dataset = None,
):
    n_folds = meta_data["n_folds"]
    seed = meta_data["seed"]
    batch_size = meta_data["batch_size"]
    cv = KFold(n_splits=n_folds, shuffle=True, random_state=seed)

    test_dataloader = (
        get_dataloader(dataset_test, batch_size, None, shuffle=False)
        if dataset_test
        else None
    )

    for fold, (train_idx, val_idx) in enumerate(cv.split(dataset_train)):
        train_dataloader = get_dataloader(
            dataset_train, batch_size, train_idx, shuffle=True
        )
        val_dataloader = get_dataloader(
            dataset_train, batch_size, val_idx, shuffle=False
        )

        run_train_test(
            train_dataloader,
            model_args,
            trainer_args,
            optimizer_args,
            metrics,
            meta_data,
            val_dataloader,
            test_dataloader,
        )



import datetime
import json
import multiprocessing
from pathlib import Path

import lightning
import numpy as np
import mlflow
import pandas as pd
import requests
import torch
import torchmetrics
from lightning.pytorch.callbacks import ModelCheckpoint

from torch.utils.data import DataLoader

from .data import load_benchmark_dataset, preprocess, PatchDataset, get_dataloader, load_and_preprocess_dataset
from .metrics import get_metrics, StableCosineSimilarity, StableKLDivergence, StableJensonShannonDivergence, \
    StableHellingerDistance
from .train import run_train_test, run_cv


def get_experiment_logger(experiment_name, run_id, tracking_uri: str | None, save_local: bool):
    if tracking_uri is not None and not save_local:
        try:
            requests.get(tracking_uri + "/health")
        except Exception:
            print("Tracking uri doesn't exist, saving locally to ./ml-runs")
            tracking_uri = None

    mlflow.enable_system_metrics_logging()
    return lightning.pytorch.loggers.MLFlowLogger(
        experiment_name=experiment_name, run_name=run_id, tracking_uri=tracking_uri
    )

def setup_experiment(experiment_cfg_path):
    with open(experiment_cfg_path, "r") as f:
        cfg = json.load(f)

    experiment_name = cfg["name"]

    run_id = cfg.get("run_id")
    if run_id is None:
        run_id = f"{experiment_name}__{int(datetime.datetime.now().timestamp())}"

    save_dir = cfg.get("out_dir")
    if save_dir is None:
        save_dir = Path(f"data/runs/{run_id}")
    save_dir.mkdir(parents=True, exist_ok=True)
    with open(save_dir / "config.json", "w") as f:
        json.dump(cfg, f)

    tracking_uri = cfg.get("mlflow_uri")
    save_local = False
    if tracking_uri is None:
        tracking_uri = str(save_dir / "mlflow_logs")
        save_local = True

    logger = get_experiment_logger(
        experiment_name, run_id, tracking_uri=tracking_uri, save_local=save_local
    )

    print(f"Running experiment '{experiment_name}' ({run_id})")
    print(f"logging results to {tracking_uri}")
    print("config:")
    print(cfg)

    return (
        cfg,
        run_id,
        save_dir,
        logger,
    )

def mulc_experiment(
    experiment_cfg_path: str = "data/config/mulc.experiment.json",
):
    cfg, run_id, save_dir, logger = setup_experiment(experiment_cfg_path)

    experiment_name = cfg["name"]
    training_args = cfg["training"]
    model_args = cfg["model"]["args"]
    optimizer_args = cfg["optimizer"]["args"]
    dataset_cfgs = cfg["datasets"]

    best_model_callback = ModelCheckpoint(
        dirpath='checkpoints/',
        filename=f'{run_id}__' + 'best-model-{epoch:02d}-{val_loss:.2f}',
        monitor='val_loss',
        mode='min',
        save_top_k=1,
        verbose=True,
    )

    trainer_args = dict(
        default_root_dir=save_dir,
        deterministic=False,
        accelerator="auto",
        max_epochs=training_args["epochs"],
        logger=logger,
        callbacks=[best_model_callback],
    )

    lightning.seed_everything(training_args["seed"])

    ds_cfg = dataset_cfgs["train"]

    patch_size = model_args["input_dimensions"][0]
    root_dir = Path(ds_cfg["base_dir"])
    df = pd.read_csv(root_dir / ds_cfg["path_df"])

    max_imgs = ds_cfg.get("max_images")
    if max_imgs is not None:
        df = df.sample(n=max_imgs)

    img_fps, mask_fps = df[["image_filepath", "mask_filepath"]].values.T

    dataset = PatchDataset(
        root_dir,
        img_fps,
        mask_fps,
        patch_size=patch_size,
        pad_value=ds_cfg["na_value"],
    )

    n = len(dataset)
    idxs = np.random.permutation(n)

    max_pixels = ds_cfg.get("max_pixels")
    if max_pixels is None:
        max_pixels = n
    n = min(n, max_pixels)

    n_train = int(n * training_args["train_percentage"])
    idxs_train = idxs[:n_train]
    idxs_val = idxs[n_train:max_pixels]

    dataloader_train = DataLoader(
        dataset=dataset,
        batch_size=training_args["batch_size"],
        sampler=torch.utils.data.sampler.SubsetRandomSampler(idxs_train),
        num_workers=training_args.get("num_workers"),
        pin_memory=True,
        persistent_workers=True,
        prefetch_factor=1,
    )
    dataloader_val = DataLoader(
        dataset=dataset,
        batch_size=training_args["batch_size"] * 16,
        sampler=torch.utils.data.sampler.SequentialSampler(idxs_val),
        num_workers=training_args.get("num_workers"),
        pin_memory=True,
        persistent_workers=True,
        prefetch_factor=1,
    )

    meta_data = {
        "experiment_name": experiment_name,
        "run_id": run_id,
        "seed": training_args["seed"],
        "datasets": ds_cfg["name"],
    }

    metrics = {
        "train": {
            "jsd": StableJensonShannonDivergence(reduction="mean"),
            "hellinger": StableHellingerDistance(reduction="mean"),
        },
        "test": {
            "jsd": StableJensonShannonDivergence(reduction="mean"),
            "hellinger": StableHellingerDistance(reduction="mean"),
        },
    }

    run_train_test(
        dataloader_train,
        model_args,
        trainer_args,
        optimizer_args,
        metrics,
        meta_data,
        dataloader_val,
        None,
        criterion=torch.nn.CrossEntropyLoss(),
    )


def cv_experiment(
    experiment_cfg_path: str = "./data/config/replication.experiment.json",
):
    cfg, run_id, save_dir, logger = setup_experiment(experiment_cfg_path)

    experiment_name = cfg["name"]
    training_args = cfg["training"]
    model_args = cfg["model"]["args"]
    optimizer_args = cfg["optimizer"]["args"]
    train_dataset_cfg = cfg["datasets"]["train"]

    lightning.seed_everything(training_args["seed"])

    dataset = load_and_preprocess_dataset(train_dataset_cfg)
    model_args["input_dimensions"] = dataset["input_dimensions"]
    model_args["num_classes"] = dataset["num_labels"]

    metrics = get_metrics(dataset["num_labels"])

    meta_data = {
        "experiment_name": experiment_name,
        "run_id": run_id,
        "seed": training_args["seed"],
        "batch_size": training_args["batch_size"],
        "n_folds": training_args.get("n_folds", 5),
        "datasets": {
            "train": dataset["name"],
            "validation": dataset["name"],
        },
    }

    trainer_args = dict(
        default_root_dir=save_dir,
        deterministic=False,
        accelerator="auto",
        max_epochs=training_args["epochs"],
        logger=logger,
    )

    run_cv(
        dataset["train"],
        model_args,
        trainer_args,
        optimizer_args,
        metrics,
        meta_data,
        dataset["test"],
    )


def ood_experiment(
    experiment_cfg_path: str = "./data/config/cross_location.experiment.json",
):
    cfg, run_id, save_dir, logger = setup_experiment(experiment_cfg_path)

    experiment_name = cfg["name"]
    training_args = cfg["training"]
    model_args = cfg["model"]["args"]
    optimizer_args = cfg["optimizer"]["args"]
    dataset_cfgs = cfg["datasets"]
    trainer_args = dict(
        default_root_dir=save_dir,
        deterministic=False,
        accelerator="auto",
        max_epochs=training_args["epochs"],
        logger=logger,
    )

    lightning.seed_everything(training_args["seed"])

    datasets = []
    for dataset_cfg in dataset_cfgs:
        datasets.append(load_benchmark_dataset(**dataset_cfg["files"]))

    batch_size = training_args["batch_size"]
    for i in range(len(datasets)):
        test_sets = datasets.copy()
        train_set = test_sets.pop(i)

        features, labels = train_set

        features_to_process = ["data_HS_LR"]
        X, y, (train_idxs, test_idxs), label_encoder, pcas = preprocess(features, labels, features_to_process=features_to_process)

        num_labels = len(label_encoder.classes_)
        model_args["input_dimensions"] = X.shape[2:]
        model_args["num_classes"] = num_labels
        metrics = get_metrics(num_labels)

        train_dataset = torch.utils.data.TensorDataset(
            torch.tensor(X[train_idxs]),
            torch.tensor(y[train_idxs])
        )
        test_dataset = torch.utils.data.TensorDataset(
            torch.tensor(X[test_idxs]),
            torch.tensor(y[test_idxs])
        )

        n = len(train_dataset)
        idxs = np.random.permutation(n)
        n_train = int(n * training_args["train_percentage"])
        idxs_train = idxs[:n_train]
        idxs_val = idxs[n_train:]

        test_datasets = [test_dataset]
        for features, labels in test_sets:
            X, y, (train_idxs, test_idxs), _, _ = preprocess(
                features,
                labels,
                features_to_process=features_to_process,
                pcas=pcas,
                label_encoder=label_encoder,
            )
            idxs = np.concatenate([train_idxs[0], test_idxs[0]])
            test_datasets.append(torch.utils.data.TensorDataset(
                torch.tensor(X[idxs]),
                torch.tensor(y[idxs]),
            ))

        train_dataloader = get_dataloader(
            train_dataset,
            batch_size=batch_size,
            shuffle=True,
            idxs=idxs_train,
        )
        val_dataloader = get_dataloader(
            train_dataset,
            batch_size=batch_size,
            shuffle=False,
            idxs=idxs_val,
        )
        test_dataloader = [
            get_dataloader(
                test_ds,
                batch_size=batch_size,
                shuffle=False,
                num_workers=min(multiprocessing.cpu_count(), 6),
            )
            for test_ds in test_datasets
        ]

        meta_data = {
            "experiment_name": experiment_name,
            "run_id": run_id,
            "seed": training_args["seed"],
            "datasets": " | ".join([ds["name"] for ds in datasets]),
        }

        run_train_test(
            train_dataloader,
            model_args,
            trainer_args,
            optimizer_args,
            metrics,
            meta_data,
            val_dataloader,
            test_dataloader,
        )



import torch

activations = {
    "gelu": torch.nn.GELU,
    "relu": torch.nn.ReLU,
}


def DepthWiseConv2d(in_channels, kernel_size, num_kernels=1, *args, **kwargs):
    """
    Creates a depthwise 2D convolutional layer.

    Args:
       in_channels (int): Number of input channels
       kernel_size (int or tuple): Size of the convolving kernel
       num_kernels (int, optional): Number of kernels per input channel. Defaults to 1
       *args: Additional positional arguments passed to Conv2d
       **kwargs: Additional keyword arguments passed to Conv2d

    Returns:
       torch.nn.Conv2d: A depthwise convolutional layer where each input channel is
       convolved with its own set of filters
    """
    return torch.nn.Conv2d(
        in_channels,
        in_channels * num_kernels,
        kernel_size=kernel_size,
        groups=in_channels,
        padding="same",
        *args,
        **kwargs,
    )


class ParallelDepthwiseConv2d(torch.nn.Module):
    """
    Implements a multi-kernel depthwise convolution module that applies multiple kernel sizes
    in parallel and adds their outputs.

    Args:
        in_channels (int): Number of input channels
        kernel_sizes (list or tuple): List of kernel sizes to use for parallel convolutions
        *args: Additional positional arguments passed to DepthWiseConv2d
        **kwargs: Additional keyword arguments passed to DepthWiseConv2d

    Forward Args:
        x (torch.Tensor): Input tensor of shape [batch_size, channels, height, width]

    Returns:
        torch.Tensor: Output tensor after applying all kernels and summing results
    """

    def __init__(self, in_channels, kernel_sizes, *args, **kwargs):
        super().__init__()
        self.kernels = torch.nn.ModuleList()
        for kernel_size in kernel_sizes:
            self.kernels.append(
                DepthWiseConv2d(in_channels, kernel_size, *args, **kwargs)
            )

    def forward(self, x):
        return torch.stack([kernel(x) for kernel in self.kernels], dim=1).sum(dim=1)


class SpatialGatedUnit(torch.nn.Module):
    """
    Implements a Spatial Gated Unit.

    This unit splits input features into two paths and modulates outputs of one path by learned interactions across
    features of the second.

    Args:
        in_features (int): Number of input features (must be even as it's split in half)
        sequence_length (int): Length of input sequence for weight matrix dimensions
        epsilon (float, optional): Range for uniform weight initialization. Defaults to 1e-3
        channel_dim (int, optional): Dimension along which to split channels. Defaults to -1

    Forward Args:
        x (torch.Tensor): Input tensor to be gated [batch_size, sequence_length, in_features]

    Returns:
        torch.Tensor: Gated output tensor with shape [batch_size, sequence_length, in_features // 2]

    See Also:
        - "Pay Attention to MLPs" by Liu et al. 2021, https://arxiv.org/abs/2105.08050
    """

    def __init__(self, in_features, sequence_length, epsilon=1e-3, channel_dim=-1):
        super().__init__()
        self.channel_dim = channel_dim
        self.epsilon = epsilon

        self.layer_norm = torch.nn.LayerNorm(in_features // 2)
        self.weights = torch.nn.Parameter(torch.empty(sequence_length, sequence_length))
        self.bias = torch.nn.Parameter(torch.ones(sequence_length, 1))

        torch.nn.init.uniform_(self.weights, -self.epsilon, self.epsilon)

    def forward(self, x):
        u, v = torch.chunk(x, 2, dim=self.channel_dim)
        v = self.layer_norm(v)
        return u * ((self.weights @ v) + self.bias)


class MLPBlock(torch.nn.Module):
    """
    A Multi-Layer Perceptron block that optionally incorporates a Spatial Gated Unit between linear layers.

    Args:
        in_features (int): Number of input features
        hidden_features (int): Number of hidden features
        sequence_length (int, optional): Required when use_sgu=True for SGU dimensions
        activation (str, optional): Name of activation function ('gelu' or 'relu'). Defaults to 'gelu'
        use_sgu (bool, optional): Whether to use Spatial Gated Unit. Defaults to False
        out_features (int, optional): Number of output features. Defaults to in_features if None
        dropout (float, optional): Dropout probability. Defaults to 0.0

    Forward Args:
        x (torch.Tensor): Input tensor [batch_size, sequence_length, in_features]

    Returns:
        torch.Tensor: Transformed output with shape [batch_size, sequence_length, out_features]

    Raises:
        ValueError: If sequence_length is None when use_sgu=True
    """

    def __init__(
        self,
        in_features,
        hidden_features,
        sequence_length=None,
        activation="gelu",
        use_sgu=False,
        out_features=None,
        dropout=0.0,
    ):
        super().__init__()
        if sequence_length is None and use_sgu:
            raise ValueError("When using SGU, sequence_length must be specified")

        if out_features is None:
            out_features = in_features

        self.activation = activations.get(activation, torch.nn.GELU)()
        self.dropout = torch.nn.Dropout(dropout)
        self.mlp1 = torch.nn.Linear(in_features, hidden_features)

        if use_sgu:
            self.sgu = SpatialGatedUnit(hidden_features, sequence_length)
            self.mlp2 = torch.nn.Linear(hidden_features // 2, out_features)
        else:
            self.sgu = torch.nn.Identity()
            self.mlp2 = torch.nn.Linear(hidden_features, out_features)

    def forward(self, x):
        x = self.mlp1(x)
        x = self.sgu(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.mlp2(x)
        x = self.dropout(x)
        return x


class MLPMixerBlock(torch.nn.Module):
    """
    MLP-Mixer block combining token and channel mixing.
    It can optionally use Spatial Gated Units in MixerBlocks.

    Args:
        in_features (int): Number of input features per token
        sequence_length (int): Number of tokens in the sequence
        hidden_features_channel (int): Hidden dimension size for channel mixing
        hidden_features_sequence (int): Hidden dimension size for token mixing
        activation (str, optional): Activation function name. Defaults to 'gelu'
        use_sgu (bool, optional): Whether to use SGU in mixing operations. Defaults to False
        dropout (float, optional): Dropout probability. Defaults to 0.0

    Forward Args:
        x (torch.Tensor): Input tensor [batch_size, sequence_length, in_features]

    Returns:
        torch.Tensor: Processed tensor with same shape as input after both mixing operations
    """

    def __init__(
        self,
        in_features,
        sequence_length,
        hidden_features_channel,
        hidden_features_sequence,
        activation="gelu",
        use_sgu=False,
        dropout=0.0,
    ):
        super().__init__()

        self.dropout = torch.nn.Dropout(dropout)
        self.layer_norm = torch.nn.LayerNorm([in_features])
        self.channel_mixer = MLPBlock(
            in_features=in_features,
            hidden_features=hidden_features_channel,
            sequence_length=sequence_length,
            activation=activation,
            dropout=dropout,
            use_sgu=use_sgu,
        )
        self.token_mixer = MLPBlock(
            in_features=sequence_length,
            hidden_features=hidden_features_sequence,
            sequence_length=in_features,
            activation=activation,
            dropout=dropout,
            use_sgu=use_sgu,
        )

    def forward(self, x):
        residual = x

        x = self.layer_norm(x).transpose(1, 2)
        x = self.token_mixer(x).transpose(1, 2)
        x = self.dropout(x)

        residual = residual + x

        x = self.layer_norm(residual)
        x = self.channel_mixer(x)
        x = self.dropout(x)
        return residual + x


def calc_n_tokens(height, width, patch_size):
    """
    Calculates the number of tokens when dividing an image into square patches.

    Args:
        height (int): Image height in pixels
        width (int): Image width in pixels
        patch_size (int): Size of each square patch

    Returns:
        int: Total number of tokens (patches)

    Note:
        Assumes the image dimensions are perfectly divisible by patch_size
    """
    return height * width // patch_size**2


class MLPMixer(torch.nn.Module):
    """
    A MLP-based vision architecture that processes images using token and channel mixing.

    This implementation splits images into patches, projects them to a token embedding space,
    and processes them through multiple MLPMixer blocks. It can optionally use Spatial Gated
    Units instead of standard MLPs.

    Args:
        image_dimensions (tuple): Input image shape as (height, width, channels)
        patch_size (int): Size of square patches to divide the image
        token_features (int): Dimension of token embedding space
        mixer_features_channel (int): Hidden dimension for channel-mixing MLPs
        mixer_features_sequence (int): Hidden dimension for token-mixing MLPs
        mixer_use_sgu (bool, optional): Whether to use SGU in mixer blocks. Defaults to False
        num_blocks (int, optional): Number of sequential mixer blocks. Defaults to 1
        activation (str, optional): Activation function type. Defaults to 'gelu'
        dropout (float, optional): Dropout probability. Defaults to 0.0

    Forward Args:
        x (torch.Tensor): Input image [batch_size, channels, height, width]

    Returns:
        torch.Tensor: Processed features [batch_size, n_tokens, token_features]

    See Also:
        - "MLP-Mixer: An all-MLP Architecture for Vision" by Tolstikhin et al. 2021, https://arxiv.org/abs/2105.01601
        - "Pay Attention to MLPs" by Liu et al. 2021, https://arxiv.org/abs/2105.08050
    """

    def __init__(
        self,
        image_dimensions,
        patch_size: int,
        token_features: int,
        mixer_features_channel: int,
        mixer_features_sequence: int,
        mixer_use_sgu=False,
        num_blocks=1,
        activation="gelu",
        dropout=0.0,
    ):
        super().__init__()
        h, w, c = image_dimensions
        self.n_tokens = calc_n_tokens(h, w, patch_size)
        self.n_channels = token_features

        self.token_embedding = torch.nn.Conv2d(
            c, token_features, kernel_size=patch_size, stride=patch_size
        )
        self.mixer_blocks = torch.nn.ModuleList()
        for _ in range(num_blocks):
            self.mixer_blocks.append(
                MLPMixerBlock(
                    in_features=self.n_channels,
                    hidden_features_channel=mixer_features_channel,
                    hidden_features_sequence=mixer_features_sequence,
                    sequence_length=self.n_tokens,
                    activation=activation,
                    use_sgu=mixer_use_sgu,
                    dropout=dropout,
                )
            )

    def forward(self, x):
        b = x.shape[0]
        x = self.token_embedding(x.moveaxis(-1, 1))
        x = x.reshape(b, self.n_channels, self.n_tokens).transpose(1, 2)
        for block in self.mixer_blocks:
            x = block(x)
        return x


class Classifier(torch.nn.Module):
    """
    Classification head that pools tokens and applies linear classification.

    Args:
        num_classes (int): Number of output classes
        in_features (int): Number of input features per token
        dropout (float, optional): Dropout probability. Defaults to 0.0

    Forward Args:
        x (torch.Tensor): Input features [batch_size, n_tokens, in_features]

    Returns:
        torch.Tensor: Classification logits [batch_size, num_classes]
    """
    def __init__(self, num_classes, in_features, dropout=0.0):
        super().__init__()
        self.avg_pool_tokens = torch.nn.AdaptiveAvgPool1d(1)
        self.dropout = torch.nn.Dropout(dropout)
        self.clf = torch.nn.Linear(in_features, num_classes)

    def forward(self, x):
        x = self.avg_pool_tokens(x.transpose(-1, -2)).transpose(-1, -2).squeeze()
        x = self.dropout(x)
        return self.clf(x)


class SGUMLPMixer(torch.nn.Module):
    """
    MLP-Mixer variant using Spatial Gated Units (SGU) during mixing and parallel depthwise convolutions with residual
    connections for preprocessing.

    This implementation allows setting the weight of the residual connection in preprocessing and optionally allows
    to make it learnable.

    Args:
        input_dimensions (tuple): Input patch dimensions as (height, width, channels)
        token_features (int): Dimension of the token embedding space
        mixer_features_channel (int): Hidden dimension for channel-mixing SGU/MLP blocks
        mixer_features_sequence (int): Hidden dimension for token-mixing SGU/MLP blocks
        mixer_use_sgu (bool, optional): Whether to use SGU blocks instead of standard MLPs. Defaults to True
        dwc_kernels (tuple, optional): Kernel sizes for parallel depthwise convolutions. Defaults to (1, 3, 5)
        use_dwc (bool, optional): Whether to use depthwise convolutions. Defaults to True.
        num_blocks (int, optional): Number of sequential mixer blocks. Defaults to 1
        activation (str, optional): Activation function for blocks. Defaults to 'gelu'
        residual_weight (float, optional): Scaling factor for residual connection. Defaults to 2
        learnable_residual (bool, optional): Whether residual weight is learnable. Defaults to False
        embedding_kernel_size (int, optional): Kernel size for token embedding conv layer. Defaults to 1
        num_classes (int, optional): Number of output classes. If 0, returns embeddings. Defaults to 0
        dropout (float, optional): Dropout rate for mixer blocks and classifier. Defaults to 0.0
        channels_first (bool, optional): Whether input has channels in first dimension. Defaults to False

    Forward Args:
        x (torch.Tensor): Input tensor of shape:
            (batch_size, channels, height, width) if channels_first=True
            (batch_size, height, width, channels) if channels_first=False

    Returns:
        torch.Tensor: Output tensor of shape:
            (batch_size, num_classes) if num_classes > 0
            (batch_size, n_tokens, token_features) if num_classes = 0
            where n_tokens = (height - embedding_kernel_size + 1)^2

    Raises:
        ValueError: If patch dimensions are not square
        ValueError: If patch height/width is not odd
        ValueError: If embedding kernel size is larger than patch size

    See Also:
        - "Spatial Gated Multi-Layer Perceptron for Land Use and Land Cover Mapping" by Jamali et al. 2024, https://github.com/aj1365/SGUMLP/blob/main/Spatial_Gated_Multi-Layer_Perceptron_for_Land_Use_and_Land_Cover_Mapping.pdf
    """

    def __init__(
        self,
        input_dimensions,
        token_features: int,
        mixer_features_channel: int,
        mixer_features_sequence: int,
        mixer_use_sgu=True,
        dwc_kernels=(1, 3, 5),
        use_dwc=True,
        num_blocks=1,
        activation="gelu",
        residual_weight=2,
        learnable_residual=False,
        embedding_kernel_size=1,
        num_classes=0,
        dropout=0.0,
        channels_first=False,
    ):
        super().__init__()
        self.input_dimensions = input_dimensions

        # validate dimensions
        patch_height, patch_width, patch_channels = input_dimensions
        if patch_height != patch_width:
            raise ValueError("Patch size must be square")
        if patch_height % 2 == 0:
            raise ValueError("Patch height must be odd")

        self.channels_first = channels_first
        self.num_classes = num_classes
        self.patch_size = patch_height
        self.embedding_kernel_size = embedding_kernel_size
        self.embedding_patch_size = patch_height - embedding_kernel_size + 1
        self.n_tokens = self.embedding_patch_size * self.embedding_patch_size
        self.n_channels = token_features
        self.dropout_rate = dropout

        self.output_dimensions = (self.n_tokens, self.n_channels) if num_classes == 0 else (num_classes,)

        self.dwc = ParallelDepthwiseConv2d(patch_channels, dwc_kernels) if use_dwc else torch.nn.Identity()
        residual_module = torch.nn.Parameter if learnable_residual else torch.nn.Buffer
        self.residual_weight = residual_module(torch.tensor(residual_weight))
        self.token_embedding = torch.nn.Conv2d(
            patch_channels,
            self.n_channels,
            kernel_size=embedding_kernel_size,
            padding="valid",
        )
        self.mixer_blocks = torch.nn.Sequential(
            *[
                MLPMixerBlock(
                    in_features=self.n_channels,
                    sequence_length=self.n_tokens,
                    hidden_features_channel=mixer_features_channel,
                    hidden_features_sequence=mixer_features_sequence,
                    activation=activation,
                    use_sgu=mixer_use_sgu,
                    dropout=self.dropout_rate,
                )
                for _ in range(num_blocks)
            ]
        )
        self.head = (
            Classifier(num_classes, self.n_channels, self.dropout_rate)
            if num_classes > 0
            else torch.nn.Identity()
        )

    def forward(self, x):
        b = x.shape[0]
        if not self.channels_first:
            x = x.moveaxis(-1, 1)
        residual = x
        x = self.dwc(x)
        x = x + (residual * self.residual_weight)
        x = self.token_embedding(x)
        x = x.reshape(b, self.n_channels, self.n_tokens).transpose(1, 2)
        x = self.mixer_blocks(x)
        x = self.head(x)
        return x






from functools import partial
from pathlib import Path

import gdown
import numpy as np
import patoolib
import rasterio
import scipy.io as sio
import torch
from PIL import Image
from numpy.lib.stride_tricks import sliding_window_view
from sklearn.decomposition import PCA
from sklearn.preprocessing import LabelEncoder
from tqdm import tqdm


def load_matrix(fp: Path | str):
    fp = Path(fp)
    return sio.loadmat(str(fp))[fp.stem]


def load_pil(fp: Path | str):
    fp = Path(fp)
    return np.array(Image.open(fp))


def load_rasterio(fp: Path | str):
    with open(fp, "rb") as f:
        with rasterio.open(f) as src:
            return src.read()


def load_image(fp: Path | str, channel_first=True):
    fp = Path(fp)
    match fp.suffix:
        case ".mat":
            img = load_matrix(fp)
        case ".tiff" | ".tif":
            img = load_rasterio(fp)
            if not channel_first:
                img = np.swapaxes(img, 0, -1)
        case _:
            img = load_pil(fp)
            if channel_first:
                img = np.swapaxes(img, 0, -1)
    return img


def load_benchmark_dataset(base_dir, feature_files, label_files, na_value=0, **kwargs):
    load = partial(load_image, channel_first=False)

    base_dir = Path(base_dir)
    feature_files = [Path(feature_file) for feature_file in feature_files]
    features = {
        file_name.stem: load(base_dir / file_name) for file_name in feature_files
    }

    if type(label_files) == str:
        label_files = [Path(label_files), None]

    if len(label_files) == 1:
        label_files = [label_files[0], None]

    file_train, file_test = label_files
    train_labels = load(base_dir / file_train)
    if file_test is not None:
        test_labels = load(base_dir / file_test)
    else:
        test_labels = np.full_like(train_labels, na_value)
    labels = {
        "train": train_labels,
        "test": test_labels,
    }
    return features, labels


class PatchDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        data_root,
        input_img_fps,
        target_img_fps,
        patch_size=9,
        pad_value=0,
        mmap_dir=None,
        mmap_max_images=None,
        mmap_init_flush_every=100,
        mmap_init_overwrite=False,
    ):
        super().__init__()
        self.data_root = Path(data_root)
        self.input_img_fps = np.asarray(input_img_fps)
        self.target_img_fps = np.asarray(target_img_fps)

        self.patch_size = patch_size
        self.pad_value = pad_value

        self._sample_img, self._sample_target = self._load_images(
            input_img_fps[0], target_img_fps[0]
        )
        if not (self._sample_img.shape[1:] == self._sample_target.shape[1:]):
            raise ValueError("Shape of input and target images do not match")

        if mmap_dir is None:
            mmap_dir = self.data_root / "mmap"

        if mmap_max_images is None:
            mmap_max_images = self.num_images

        self.mmap_max_images = mmap_max_images
        self.mmap_init_flush_every = mmap_init_flush_every
        self.mmap_init_overwrite = mmap_init_overwrite
        if self.mmap_max_images > 0:
            self.mmap_dir = Path(mmap_dir)
            (
                self._mmap_images,
                self._mmap_targets,
                self._mmap_images_path,
                self._mmap_targets.path,
            ) = self._init_mmaps(
                flush_every=mmap_init_flush_every, overwrite=mmap_init_overwrite
            )
        else:
            (
                self._mmap_images,
                self._mmap_targets,
                self._mmap_images_path,
                self._mmap_targets.path,
                self.mmap_dir,
            ) = (
                None,
                None,
                None,
                None,
                None,
            )

    @property
    def pad_size(self):
        return self.patch_size // 2

    @property
    def num_pixels(self):
        return self.height * self.width

    @property
    def num_images(self):
        return self.input_img_fps.shape[0]

    @property
    def height(self):
        return self._sample_img.shape[1] - 2 * self.pad_size

    @property
    def width(self):
        return self._sample_img.shape[2] - 2 * self.pad_size

    @property
    def image_channels(self):
        return self._sample_img.shape[0]

    @property
    def target_channels(self):
        return self._sample_target.shape[0]

    @property
    def image_dimensions(self):
        return self.image_channels, self.height, self.width

    @property
    def target_dimensions(self):
        return self.target_channels, self.height, self.width

    def __len__(self):
        return self.num_images * self.num_pixels

    def _init_mmaps(self, flush_every=100, overwrite=False):
        map_params = [
            (
                "input.mmap",
                (self.mmap_max_images, *self._sample_img.shape),
                self._sample_img.dtype,
            ),
            (
                "target.mmap",
                (self.mmap_max_images, *self._sample_target.shape),
                self._sample_target.dtype,
            ),
        ]
        self.mmap_dir.mkdir(parents=True, exist_ok=True)

        maps = []
        paths = []
        for filename, shape, dtype in map_params:
            filepath = self.mmap_dir / filename
            paths.append(filepath)
            if not filepath.exists():
                arr = np.memmap(filepath, dtype=dtype, mode="w+", shape=shape)
                arr.flush()

            else:
                arr = np.memmap(filepath, dtype=dtype, mode="r+", shape=shape)
            maps.append(arr)

        mmap_images, mmap_targets = maps

        if not all([path.exists() for path in paths]) or overwrite:
            for i in tqdm(
                range(self.mmap_max_images),
                total=self.mmap_max_images,
                unit="image pairs",
            ):
                img, target = self.load_image_and_target(i)
                mmap_images[i] = img
                mmap_targets[i] = target

                if i % flush_every == 0:
                    mmap_images.flush()
                    mmap_targets.flush()

            mmap_images.flush()
            mmap_targets.flush()

        path_mmap_img, path_mmap_target = paths
        return mmap_images, mmap_targets, path_mmap_img, path_mmap_target

    def _get_paths(self, i):
        return (self.input_img_fps[i], self.target_img_fps[i])

    def _load_images(self, image_fp: str, target_img_fp: str):
        image_fp = self.data_root / image_fp
        target_img_fp = self.data_root / target_img_fp
        img = load_rasterio(image_fp)
        target = load_rasterio(target_img_fp)

        pad_args = dict(
            pad_width=(
                (0, 0),
                (self.pad_size, self.pad_size),
                (self.pad_size, self.pad_size),
            ),
            mode="constant",
            constant_values=self.pad_value,
        )
        return np.pad(img, **pad_args), np.pad(target, **pad_args)

    def load_image_and_target(self, i):
        image_fp, target_img_fp = self._get_paths(i)
        return self._load_images(image_fp, target_img_fp)

    def get_image_and_target(self, i: int):
        if i < self.mmap_max_images:
            img = self._mmap_images[i].copy()
            target = self._mmap_targets[i].copy()
            return img, target
        return self.load_image_and_target(i)

    def __getitem__(self, idx):
        i = idx // self.num_pixels
        patch_idx = idx % self.num_pixels
        h = patch_idx // self.height
        h_ = h + self.patch_size
        w = patch_idx % self.width
        w_ = w + self.patch_size

        if i < self.mmap_max_images:
            patch = self._mmap_images[i, :, h:h_, w:w_].copy()
            mask = self._mmap_targets[i, :, h, w].copy()
        else:
            img, target = self.get_image_and_target(i)
            patch = img[:, h:h_, w:w_]
            mask = target[:, h, w]

        return torch.tensor(patch).float(), torch.tensor(mask).float()


def _is_valid(a, na_value):
    return a != na_value


def patchify(
    image: np.ndarray,
    patch_size: int = 9,
    pad_value=0,
):
    if patch_size % 2 == 0:
        raise ValueError("patch_size must be odd")
    offset = patch_size // 2

    image = np.array(image, copy=True)
    if image.ndim < 3:
        raise ValueError("image must have at least 3 dimensions (H, W, C) or 4 dimensions (N, H, W, C)")

    if image.ndim == 3:
        image = np.expand_dims(image, 0)

    # configure sliding windows
    pad_width = [(0, 0), (offset, offset), (offset, offset)]
    window_shape = [1, patch_size, patch_size]

    # include feature dimensions
    n_feature_dims = image.ndim - 3
    pad_width.extend([(0, 0)] * n_feature_dims)
    window_shape.extend(image.shape[3:])

    # extract patches
    image = np.pad(
        image, pad_width=pad_width, mode="constant", constant_values=pad_value
    )
    image = sliding_window_view(image, window_shape=window_shape).squeeze()
    return image


def reduce_dimensions(
    image: np.ndarray, num_components=15, channel_dim=-1, train_mask=None, pca=None
):
    channels_last = channel_dim == -1 or channel_dim == (image.ndim - 1)

    if not channels_last:
        image = np.swapaxes(image, channel_dim, -1)
        if train_mask is not None and train_mask.shape == image.shape:
            train_mask = np.swapaxes(train_mask, channel_dim, -1)

    image_dimensions = image.shape[:-1]
    c = image.shape[-1]

    if train_mask is not None:
        X = image[train_mask]
    else:
        X = image

    X = X.reshape(-1, c)

    if pca is None:
        pca = PCA(n_components=num_components, whiten=True).fit(X)

    X = pca.transform(image.reshape(-1, c)).reshape(*image_dimensions, num_components)

    if not channels_last:
        X = np.swapaxes(X, -1, channel_dim)

    return X, pca


def preprocess(
    features: dict[str, np.ndarray],
    labels: dict[str, np.ndarray],
    image_dtype=np.float32,
    features_to_process: list[str] | None = None,
    patch_size: int = 9,
    na_value: int = 0,
    num_components=15,
    channel_dim=-1,
    pcas=None,
    label_encoder=None,
):
    image_dimensions = labels["train"].shape[:2]
    num_pixels = np.prod(image_dimensions)

    valid_labels = []
    idxs = []
    for split, targets in labels.items():
        mask = np.asarray(targets != na_value)
        valid_labels.append(targets[mask])
        idxs.append(np.where(mask.reshape(num_pixels)))

    if label_encoder is None:
        label_encoder = LabelEncoder().fit(np.concatenate(valid_labels, axis=0).ravel())

    y_train, y_test = [label_encoder.transform(y) for y in valid_labels]
    idxs_train, idxs_test = idxs
    y = np.full(num_pixels, -1, dtype=int)
    y[idxs_train] = y_train
    y[idxs_test] = y_test

    if features_to_process is None:
        features_to_process = []
    features_to_process = set(features_to_process)

    if pcas is None:
        pcas = {}

    X = []
    for name, data in features.items():
        data = data.reshape(num_pixels, -1)
        if name in features_to_process:
            pca = pcas.get(name)
            data, pca = reduce_dimensions(
                data,
                num_components,
                channel_dim=channel_dim,
                train_mask=idxs_train,
                pca=pca,
            )
            pcas[name] = pca
        X.append(data)

    X = np.concatenate(X, axis=channel_dim)
    channels = X.shape[-1]
    X = X.reshape(-1, *image_dimensions, channels)
    X = patchify(X, patch_size=patch_size, pad_value=na_value).reshape(-1, patch_size, patch_size, channels)

    return (X.astype(image_dtype), y, (idxs_train, idxs_test), label_encoder, pcas)


def get_dataloader(dataset, batch_size, idxs=None, shuffle=False, num_workers=1):
    sampler = None
    if idxs is not None:
        sampler = torch.utils.data.SubsetRandomSampler(idxs)
        shuffle = None
    return torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        batch_size=batch_size,
        shuffle=shuffle,
        sampler=sampler,
    )


def to_torch_dataset(X, y):
    return torch.utils.data.TensorDataset(torch.from_numpy(X), torch.from_numpy(y))


def load_and_preprocess_dataset(dataset_cfg: dict, pcas=None):
    dataset = load_benchmark_dataset(
        **dataset_cfg["files"], na_value=dataset_cfg.get("na_value", 0)
    )
    X, y, (train_idx, test_idx), label_encoder, pcas = preprocess(
        *dataset, **dataset_cfg["preprocessing"], pcas=pcas
    )
    n, p, _, c = X.shape
    X_train = X[train_idx]
    y_train = y[train_idx]
    X_test = X[test_idx]
    y_test = y[test_idx]
    dataset_train = torch.utils.data.TensorDataset(
        torch.from_numpy(X_train), torch.from_numpy(y_train)
    )
    dataset_test = torch.utils.data.TensorDataset(
        torch.from_numpy(X_test), torch.from_numpy(y_test)
    )

    return {
        "name": dataset_cfg["name"],
        "train": dataset_train,
        "test": dataset_test,
        "num_labels": len(label_encoder.classes_),
        "input_dimensions": X_train.shape[1:],
        "pcas": pcas,
        "label_encoder": label_encoder,
    }


def download_benchmark_datasets(data_dir="data", overwrite=False):
    data_path = Path(data_dir)
    datasets_dir = data_path / "Datasets"
    if datasets_dir.exists() and not overwrite:
        print("Dataset already exists, skipping download. If you want to overwrite it, pass overwrite=True or delete it.")
        return

    data_path.mkdir(parents=True, exist_ok=True)
    file_id = "1dLJJrNJpQoQeDHybs37iGxmrSU6aP2xv"  # https://drive.usercontent.google.com/download?id=1dLJJrNJpQoQeDHybs37iGxmrSU6aP2xv&export=download
    file_path = data_path / (file_id + ".rar")
    try:
        gdown.download(id=file_id, output=str(file_path), quiet=False)
        patoolib.extract_archive(str(file_path), outdir=str(data_path))
    finally:
        file_path.unlink(missing_ok=True)
